"""
Speak a string of text or speak the contents of a text file.

Currently, these routines expect VLC to be installed.

Example::
    from dt_tools.os.sound import Accent, Sound

    obj = Sound()
    obj.speak('This is a test')
    obj.speak('This is a test, with an australian accent.', accent=Accent.Australia)

ToDo:

    Update to be cross platform without relying on VLC

"""
import os
import pathlib
import textwrap
from time import sleep

from gtts import gTTS
from loguru import logger as LOGGER

from dt_tools.os.os_helper import OSHelper as helper
from enum import Enum

class Accent(Enum):
    """Accent codes for speaking"""
    Australia = "com.au"
    UnitedKingdom = "co.uk"
    UnitedStates = "us"
    Canada = "ca"
    India = "co.in"
    Ireland = "ie"
    SouthAfrica = "co.za"
    Nigeria = "com.ng"


class Sound():
    """
    Class to speak a string of text (or contents of a text file).

    This class relies on VLC being installed, it works on both Windows and Linux.

    Raises:
        FileNotFoundError: If file is not found.

    """
    _instance = None
    _is_speaking = False

    def __new__(cls):
        # Make this class a singleton
        if cls._instance is None:
            LOGGER.debug('creating sound class')
            cls._instance = super(Sound, cls).__new__(cls)
        return cls._instance
    
    def __init__(self):
        if helper.is_windows():
            start_path = pathlib.Path(os.environ['ProgramFiles'])
            exe = "vlc.exe"
        else:
            start_path = pathlib.Path('/usr/bin')
            exe = 'cvlc'
        
        self._VLC = self._file_location(start_path, exe)
        if self._VLC is None:
            raise FileNotFoundError('Unable to locate VLC module')
    
    def speak(self, in_token: str, speed: float = 1.0, accent: Accent = Accent.UnitedStates) -> int:
        """
        Speak the text string or contents of the file

        Args:
            in_token (str): File or string of text to be spoken
            speed (float, optional): Speed (cadence) of voice. Defaults to 1.0.
            accent (Accent, optional): Accent of speaker. Defaults to Accent.UnitedStates.

        Returns:
            int: 0 if successful else non-zero
        """
        while self._is_speaking:
            LOGGER.debug('waiting..')   
            sleep(1)
        self._is_speaking = True
        check_file = pathlib.Path(in_token)
        try:
            is_file = check_file.is_file()
        except OSError:
            is_file = False
        text = check_file.read_text() if is_file else in_token

        sound_file = './da_sound.mp3'
        # tld top level domain for English
        # com.au (Australian), co.uk (United Kingdom), us (United States),    ca (Canada), 
        # co.in (India),       ie (Ireland),           co.za (South Africa),  com.ng (Nigeria)
        tts_obj = gTTS(text=text, lang='en', tld=accent.value, slow=False)
        tts_obj.save(sound_file)
        
        display_text = textwrap.wrap(text=text, width=100, initial_indent='- Speak: ', subsequent_indent='         ')
        for line in display_text:
            LOGGER.debug(line)
        ret = self._play(sound_file, speed)
        self._is_speaking = False

        return ret

    def play(self, sound_file: str, speed: float = 1.0) -> int:
        """
        Play a sound file.

        Args:
            sound_file (str): Filename
            speed (float, optional): Speed (cadence) of voice. Defaults to 1.0.

        Returns:
            int: 0 if successful else non-zero
        """
        while self._is_speaking:
            LOGGER.debug('waiting..')
            sleep(1)
        self._is_speaking = True
        result = self._play(sound_file, speed)
        self._is_speaking = False
        return result
    
    def _play(self, sound_file: str, speed: float = 1.0) -> int:
        '''Play the sound file'''
        check_file = pathlib.Path(sound_file)
        if not check_file.is_file():
            msg = f'Sorry, sound file {sound_file} does not exist.'
            LOGGER.warning(msg)
            return -1
        
        LOGGER.debug(f'Playing file: {self._VLC} {sound_file}')
        if helper.is_windows():
            ret = os.system(f'"{self._VLC}" --intf dummy --rate {speed} --play-and-exit {sound_file}')
        else:
            ret = os.system(f'{self._VLC} --rate {speed} --play-and-exit {sound_file}')

        return  ret
    
    def _file_location(self, search_path: str, target: str) -> str:
        vlc_loc: str = None
        for filepath in search_path.rglob(target):
            vlc_loc = filepath
            break

        return vlc_loc
    
if __name__ == "__main__":
    obj = Sound()
    obj.speak('This is a test')
    obj.speak('This is a test, with an australian accent.', accent=Accent.Australia)
    obj.speak('This is a test, with a british accent.', accent=Accent.UnitedKingdom)
    obj.speak('This is a test, with a canadian accent.', accent=Accent.Canada)
    obj.speak('This is a test, with a indian accent.', accent=Accent.India)
    obj.speak('This is a test, with a irish accent.', accent=Accent.Ireland)
    obj.speak('This is a test, with a nigerian accent.', accent=Accent.Nigeria)
    obj.speak('This is a test, with a south african accent.', accent=Accent.SouthAfrica)
    obj.speak('C:/Workspace/ws_photo/da-photo/research/test.txt')



import functools
import logging
import logging.handlers
import pathlib
import sys


DEFAULT_FILE_LOGFMT = "%(asctime)s [%(module)-15s] [%(levelname)-8s] %(message)s"
"""For file logging, format- timestamp \|level\|method name\|lineno\|message"""

DEFAULT_CONSOLE_LOGFMT = "%(message)s"
"""For console logging, format- message"""

DEFAULT_DATETIMEFMT = " %Y/%m/%d %I:%M:%S"
"""Default Time Format"""

logging.TRACE = 5
logging._levelToName[logging.TRACE] = "TRACE"
logging._nameToLevel["TRACE"] = logging.TRACE
def trace(self, msg, *args, **kwargs):
    """
    Log 'msg % args' with severity 'TRACE'.

    To pass exception information, use the keyword argument exc_info with
    a true value, e.g.

    logger.trace("Houston, we have a %s", "interesting problem", exc_info=1)
    """
    if self.isEnabledFor(logging.TRACE):
        self._log(logging.TRACE, msg, args, **kwargs)

logging.Logger.trace = trace
# TODO:
#   - Loglevels - TRACE, DEBUG, INFO, ERROR, CRITICAL

def configure_logger(logger_name: str, log_target = sys.stderr, log_level: str = "INFO", log_format: str = None, **kwargs) -> logging.Logger:
    """
    Configure logger via python logging.

    Parameters:
        logger_name: Name of logger
        log_target: defaults to stderr, but can supply filename as well
        log_level : TRACE|DEBUG|INFO(dflt)|ERROR|CRITICAL
        log_format: format for output log line
        other     : keyword args related to loguru logger.add() function

    Returns:
        logger_handle: Logger object
    """
    logger = None
    log_file = None
    
    print(f'logger_name: {logger_name}')
    if isinstance(log_target, str):
        log_file = pathlib.Path(log_target)
        if not log_file.parent.exists():
            print(f'ERROR: cannot create log file: {log_file}')
            print('       Path does not exist.')
            log_file = None

    if not logging.getLogger().hasHandlers():
        # No ROOT logger setup, set basicConfig
        print('no root logger, set basicConfig')
        logging.basicConfig(format=log_format, datefmt=DEFAULT_DATETIMEFMT, level=logging.ERROR)      

    # Create a custom logger (either Console or File)
    logger = logging.getLogger(logger_name)
    logger.propagate = False
    logger.setLevel(log_level)

    # Remove any existing handlers
    # if logger.hasHandlers():
    #     for handler in logger.handlers:
    #         logger.removeHandler(handler)

    if log_file is None:
        # Console
        print('creating console logger')
        l_handle = logging.StreamHandler()
        l_line_format = log_format if log_format is not None else DEFAULT_CONSOLE_LOGFMT
        l_handle.name = "console"
    else:
        # File logger
        print('creating file logger')
        l_handle = logging.FileHandler(log_file)
        l_line_format = log_format if log_format is not None else DEFAULT_FILE_LOGFMT
        l_handle.name = "file"

    l_handle.setLevel(log_level)
    l_format = logging.Formatter(fmt=l_line_format, datefmt=DEFAULT_DATETIMEFMT)
    l_handle.setFormatter(l_format)
    logger.addHandler(l_handle)
    print(f'returning {logger.name}')
    return logger



def logger_wraps(*, entry=True, exit=True, level="DEBUG"):
    """
    BROKEN!! function decorator wrapper to log entry and exit

    When decorator enabled, messages will automatically be included in the the log:
    Example::    

        @logger_wraps()
        def foo(a, b, c):
            logger.info("Inside the function")
            return a * b * c 

    Output:
    > 
    """
    def wrapper(func):
        # THIS DOES NOT WORK, needs to be re-addressed.
        name = func.__name__

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            # logger_ = logger.opt(depth=1)
            logger_ = logging.root
            log_level = level if isinstance(level, int) else logging._nameToLevel[level]
            if entry:
                logger_.log(log_level, "Entering '{}' (args={}, kwargs={})", name, args, kwargs)
            result = func(*args, **kwargs)
            if exit:
                logger_.log(log_level, "Exiting '{}' (result={})", name, result)
            return result

        return wrapped

    return wrapper


# def test(con_logger, file_logger):
#     for lvl in ["TRACE", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
#         lvl_no = logging._nameToLevel[lvl]
#         con_logger.log(lvl_no, f"This is a console {lvl} message.")
#         file_logger.log(lvl_no, f"This is a file {lvl} message.")
#     con_logger.trace('This is a console test for trace method')
#     file_logger.trace('This is a File test for trace method')

if __name__ == "__main__":
    print('test')
    conLog = configure_logger(logger_name='consoleLog', log_level="TRACE")
    fileLog = configure_logger(logger_name='fileLog', log_target='./test.log', log_level="TRACE")
    for lvl in ["TRACE", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
        lvl_no = logging._nameToLevel[lvl]
        conLog.log(lvl_no, f"This is a console {lvl} message.")
        fileLog.log(lvl_no, f"This is a file {lvl} message.")
    conLog.trace('This is a console test for trace method')
    fileLog.trace('This is a File test for trace method')



import dt_tools.logger.logging_helper as lh
import datetime
import time
import random

from loguru import logger as LOGGER

def demo():
    test1_log = "./test1.log"
    test2_log = "./test2.log"
    rotation=datetime.timedelta(seconds=10)
    retention= 5

    lh.configure_logger()
    LOGGER.info('')
    LOGGER.info('dt_misc_logging_demo')
    LOGGER.info('-'*80)
    LOGGER.info('')
    LOGGER.info('Begin demo...')
    lh.configure_logger(log_target=test1_log, log_level="DEBUG")
    lh.configure_logger(log_target=test2_log, log_level="TRACE",
                                    retention=retention, rotation=rotation)
    LOGGER.info('- 30 message with random log levels will be sent to the logger.')
    LOGGER.info('- Depending on configuration, each message will be routed to the appropriate logger(s)')
    LOGGER.info('')
    LOGGER.info('Logger configuration:')
    LOGGER.info('  Console   : CRITICAL, ERROR, WARNING, INFO')
    LOGGER.info('  Test1.log : CRITICAL, ERROR, WARNING, INFO, DEBUG')
    LOGGER.info('  Test2.log : CRITICAL, ERROR, WARNING, INFO, DEBUG, TRACE')
    LOGGER.info('')
    LOGGER.info('NOTE: ')
    LOGGER.info('- The Console will ONLY get INFO level and above.')
    LOGGER.info(f'- The {test1_log} file will get DEBUG level and above.')
    LOGGER.info(f'- The {test2_log} file will get TRACE level and above.')
    LOGGER.info(f'- The {test2_log} file is set to rotate every 10 seconds and have 5 total versions.')
    time.sleep(3)
    LOGGER.info('')
    LOGGER.trace('This TRACE message should ONLY print in test2.log')
    LOGGER.debug('This DEBUG message should print in test1.log and test2.log')
    LOGGER.info('This INFO message should print on console and test1/2 log files')
    for i in range(30):
        log_level = random.choice(['TRACE','DEBUG','INFO','WARNING','ERROR','CRITICAL'])
        LOGGER.log(log_level, f'message {i:2} {log_level}')
        time.sleep(1)
    LOGGER.info('')
    LOGGER.info('demo complete.')

if __name__ == "__main__":
    demo()



from loguru import logger as LOGGER

import dt_tools.logger.logging_helper as lh
from dt_tools.os.os_helper import OSHelper

def demo():
    LOGGER.info('')
    LOGGER.info('dt_misc_os_demo')
    LOGGER.info('-'*80)
    LOGGER.info('')
    LOGGER.info('Determine OS:')
    LOGGER.info('-'*40)
    if OSHelper.is_raspberrypi():
        LOGGER.info('- Running on Raspberry Pi')
    if OSHelper.is_linux():
        LOGGER.success('- Running on Linux')
    if OSHelper.is_windows():
        LOGGER.success('- Running on Windows')

    LOGGER.info('')
    LOGGER.info('Check if Executable is available:')
    LOGGER.info('-'*40)
    for pgm in ['gitx', 'git']:
        exe = OSHelper.is_executable_available(pgm)
        LOGGER.log("ERROR" if exe is None else "SUCCESS", f'  {pgm:8} : {exe}')

    if OSHelper.is_windows():
        LOGGER.info('')
        input('Press ENTER to Continue for Admin Check')
        LOGGER.info('')
        LOGGER.info('Admin Check:')
        LOGGER.info('-'*40)
        if OSHelper.is_windows_admin():
            LOGGER.info('')
            LOGGER.info('  ****************************************')
            LOGGER.info('  ** Windows admin privileges in effect **')
            LOGGER.info('  ****************************************')
            LOGGER.info('')
            input('Press Enter to continue...')        
        else:
            LOGGER.warning('  Not Admin, Elevate privileges')
            if OSHelper.elevate_to_admin():
                LOGGER.info('  - New prompt Shelled as admin')
            else:
                LOGGER.error('  - Unable to elevate. (User cancelled or error)')

        LOGGER.info('')
        LOGGER.info('Demo commplete.')            

if __name__ == "__main__":
    lh.configure_logger(log_format=lh.DEFAULT_CONSOLE_LOGFMT, log_level="INFO")
    demo()



"""
Logging helper methods for loguru
"""
import functools
import logging
import sys

from loguru import logger

DEFAULT_FILE_LOGFMT = "<green>{time:MM/DD/YY HH:mm:ss}</green> |<level>{level: <8}</level>|<cyan>{name:10}</cyan>|<cyan>{line:3}</cyan>| <level>{message}</level>"
"""For file logging, format- timestamp \|level\|method name\|lineno\|message"""

DEFAULT_CONSOLE_LOGFMT = "<level>{message}</level>"
"""For console logging, format- message"""

def configure_logger(log_target = sys.stderr, log_level: str = "INFO", log_format: str = None, log_handle: int = 0, **kwargs) -> int:
    """
    Configure logger via loguru.

     - should be done once for each logger (console, file,..)
     - if reconfiguring a logger, pass the log_handle
    
    Parameters:
        log_target: defaults to stderr, but can supply filename as well
        log_level : TRACE|DEBUG|INFO(dflt)|ERROR|CRITICAL
        log_format: format for output log line
        log_handle: handle of log being re-initialized.
        other     : keyword args related to loguru logger.add() function

    Example::

        import dt_tools.logger.logging_helper as lh
        from loguru import logger as LOGGER

        log_file = './mylog.log'

        f_handle = lh.configure_logger(log_target=log_file, log_level="DEBUG")
        c_handle - lh.configure_logger()

        LOGGER.debug('this should only show up in file logger')
        LOGGER.info('this should show up in file logger and console')

    Returns:
        logger_handle_id: integer representing logger handle
    """
    try:
        logger.remove(log_handle)
        # attempt to include all python loggers
        logging.basicConfig(handlers=[_InterceptHandler()], level=0, force=True)
    except:  # noqa: E722
        pass
    
    if not log_format:
        # Set format based on type of logger (file vs console)
        if isinstance(log_target, str):
            log_format = DEFAULT_FILE_LOGFMT
        else:
            log_format = DEFAULT_CONSOLE_LOGFMT

    hndl = logger.add(sink=log_target, level=log_level, format=log_format, **kwargs)

    return hndl

def logger_wraps(*, entry=True, exit=True, level="DEBUG"):
    """
    function decorator wrapper to log entry and exit

    When decorator enabled, messages will automatically be included in the the log:
    Example::    

        @logger_wraps()
        def foo(a, b, c):
            logger.info("Inside the function")
            return a * b * c 
    """
    def wrapper(func):
        name = func.__name__

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            logger_ = logger.opt(depth=1)
            if entry:
                logger_.log(level, "Entering '{}' (args={}, kwargs={})", name, args, kwargs)
            result = func(*args, **kwargs)
            if exit:
                logger_.log(level, "Exiting '{}' (result={})", name, result)
            return result

        return wrapped

    return wrapper

class _InterceptHandler(logging.Handler):
    def emit(self, record):
        # Get corresponding Loguru level if it exists.
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message.
        frame, depth = sys._getframe(6), 6
        while frame and frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())
        
@logger_wraps(level='INFO')
def _print_log_level_definitions():
    print(logger.level("TRACE"))
    print(logger.level("DEBUG"))
    print(logger.level("INFO"))
    print(logger.level("SUCCESS"))
    print(logger.level("WARNING"))
    print(logger.level("ERROR"))
    print(logger.level("CRITICAL"))


if __name__ == "__main__":
    import dt_tools.cli.dt_misc_logging_demo as module
     
    module.demo()



import inspect
import pathlib
from datetime import datetime as dt
from importlib.metadata import Distribution, distributions, version
from typing import Tuple, Union

from loguru import logger as LOGGER


class ProjectHelper:
    _max_depth = 4

    @staticmethod
    def _search_down_tree(filename: str, start_path: str, depth: int = _max_depth) -> Union[pathlib.Path, None]:
        """
        Search directory tree towards root for filename starting at start path.

        Limit seach to depth number of directories.

        Args:
            filename (str): target_filename.
            start_path (str): path to begin search.
            depth (int, optional): Number of directories to travers (towards root). Defaults to _max_depth.

        Returns:
            str: Full filename/path or None.
        """
        LOGGER.debug(f'  Search for {filename}')
        result_file: pathlib.Path = None
        cur_depth = 0
        traverse_path = pathlib.Path(start_path)
        pattern = f"**/{filename}"
        while result_file is None and cur_depth < depth:
            cur_depth += 1
            file_list = list(traverse_path.glob(pattern))
            LOGGER.debug(f'  - directory: {str(traverse_path)}')
            LOGGER.debug(f'    file_list: {file_list}')
            if len(file_list) == 1:
                result_file = file_list[0]
                LOGGER.debug(f'  FOUND: {result_file}')
            else:
                traverse_path = traverse_path.parent
        return result_file

    @staticmethod
    def _check_metadata(target_name: str) -> Tuple[str,str]:
        ver = None
        determined_from = None
        try:
            LOGGER.debug('Try import.metadata')
            ver = version(target_name)
            determined_from = "importlib.metadata"
        except:  # noqa: E722
            LOGGER.debug('- Not found in metadata')


        return ver, determined_from

    @staticmethod
    def _check_toml(project_name: str, calling_module: str) -> Tuple[str,str]:
        LOGGER.debug('Try pyproject.toml')
        ver = None
        LOGGER.debug(project_name)
        determined_from = None
        target_file = ProjectHelper._search_down_tree("pyproject.toml", pathlib.Path(calling_module).parent)
        # LOGGER.warning(target_file)
        if target_file is None:
            LOGGER.debug('- unable to locate pyproject.toml')
        else:
            buff = target_file.read_text(encoding='utf-8').splitlines()
            proj_name = ""
            name_list = [x for x in buff if x.startswith('name')]
            # LOGGER.warning(name_list)
            if len(name_list) >= 1:
                token = name_list[0].split('=')[1].strip()
                proj_name = token.replace('"',"").replace("'","")
                # LOGGER.warning(f'token: {token}  proj_name: {proj_name}')
            if project_name != proj_name:
                LOGGER.debug(f'- Requested project name {project_name} does not match pyproject.toml project name {proj_name}')
            else:
                ver_line = [x for x in buff if x.startswith('version')]
                LOGGER.debug(f'ver_line: {ver_line}')
                if len(ver_line) == 1:
                    ver = ver_line[0].split('=')[1].replace('"','').replace("'",'').strip()
                    determined_from = "pyproject.toml"
                LOGGER.debug('- Identified via pyproject.toml')
        return ver, determined_from
    
    @staticmethod
    def _check_call_stack(root_path: str, python_file: str) -> Tuple[str, str]:
        LOGGER.debug('Try python file')
        ver = None
        determined_from = None
        if not python_file.endswith('.py'):
            LOGGER.debug(f'  Passed file: {python_file} does not appear to be python file.')
            return ver, determined_from
        
        LOGGER.debug(f'- python file: {python_file}  root_path: {root_path}')
        file_list = list(pathlib.Path(root_path).glob(f"**/{python_file}"))
        if len(file_list) == 0:
            LOGGER.debug(f'  Unable to locate python file: {python_file}')
            return ver, determined_from
        LOGGER.debug(f'  Python file abs: {file_list[0]}')
        file_list = list(pathlib.Path(root_path).glob('**/*.py'))
        ver_date = dt(2000,1,1,0,0,0,0)
        LOGGER.debug(f'  File list: {file_list}')
        for file_nm in file_list:
            if dt.fromtimestamp(file_nm.stat().st_mtime) > ver_date:
                ver_date = dt.fromtimestamp(file_nm.stat().st_mtime)
                ver_file = file_nm
            ver_date = max(ver_date, dt.fromtimestamp(file_nm.stat().st_mtime))
        ver = f'{ver_date.year}.{ver_date.month}.{ver_date.day}'    
        determined_from = f'File date: {str(ver_date)} {ver_file}'
    
        return ver, determined_from
    
    @staticmethod
    def determine_version(target_name: str, identify_src: bool = False) -> Union[str, Tuple[str, str]]:
        """
        Retrieve project version for distribution (or running codebase)

        Version is determined by:  
        - check importlib metadata
        - look for pyproject.toml  
        - scanning calling stack root file newest python module  

        Args:
            distrib_name (str): Package distribution name.
              If distrib_name not found, version will be determined from pyproject.toml (if found) or
              from the newest .py file in the stack path starting at the calling program.
            identify_src (bool, otional): Return a string indicating how the version was determined. Defaults to false

        Returns:
            Union[str,Tuple[str,str]: version or version, source.
            - version is in format major.minor.patch or YYYY.MM.DD.
            - source is one of 'importlib.metadata', 'pyproject.toml' or source filename.
        """
        if not isinstance(target_name, str):
            raise ValueError(f'Invalid target name (must be str) in determine_version: {target_name}')
        
        LOGGER.debug(f'determine_version for {target_name}')
        ver = None
        root_idx = len(inspect.stack()) - 1
        caller = pathlib.Path(inspect.stack()[root_idx].filename)
        determined_from = None
        ver, determined_from = ProjectHelper._check_metadata(target_name)
        if ver is None:
            ver, determined_from = ProjectHelper._check_toml(target_name, caller)

        if ver is None:
            python_file = f'{target_name}.py'
            ver, determined_from = ProjectHelper._check_call_stack(root_path=caller.parent, python_file=python_file)
            if ver is None:
                root_idx = 0 # len(inspect.stack()) - 2
                caller = pathlib.Path(inspect.stack()[root_idx].filename)
                ver, determined_from = ProjectHelper._check_call_stack(root_path=caller.parent, python_file=python_file)

        if identify_src:
            return (ver, determined_from)
        return ver

    @staticmethod
    def installed_packages() -> dict:
        """
        Return a dictionary describing installed packages.

        Returns:
            dict: in format {"pkg_name1": "pkg_version", "pkg_name2": "pkg_version", ...}
        """
        package: Distribution = None
        package_dict: dict = {}
        for package in distributions():
            package_dict[package.name] = package.version
        
        return package_dict
    

    


from loguru import logger as LOGGER

import dt_tools.logger.logging_helper as lh
from dt_tools.os.project_helper import ProjectHelper


def demo():
    LOGGER.info('-'*80)
    LOGGER.info('dt_misc_project_helper_demo')
    LOGGER.info('-'*80)
    LOGGER.info('')

    LOGGER.info('Determine versions:')
    LOGGER.info(f'- Distribution dt-misc        : {ProjectHelper.determine_version("dt-misc", identify_src=True)}')
    LOGGER.info(f'- File:        project_helper : {ProjectHelper.determine_version("project_helper", identify_src=True)}')
    LOGGER.info('')
    LOGGER.info('Installed Packages:')
    LOGGER.info('  Package                        Version')
    LOGGER.info('  ------------------------------ ---------')
    for package, ver in ProjectHelper.installed_packages().items():
        print(f'  {package:30} {ver}')

    LOGGER.info('')
    LOGGER.info('Demo commplete.')            

if __name__ == "__main__":
    lh.configure_logger(log_format=lh.DEFAULT_CONSOLE_LOGFMT, log_level="INFO")
    demo()



from types import SimpleNamespace

"""
General Helper Routines

- ObjectHelper
- StringHelper

"""
# =================================================================================================
class ObjectHelper:

    @classmethod
    def dict_to_obj(cls, in_dict: dict):
        """
        Convert a dictionary to an object

        Args:
            in_dict (dict): Input dictionary

        Returns:
            dict or object: Object representation of dictionary, or Object if not a dictionary

        Raises:
            TypeError if in_dict is NOT a dictionary.            
        """
        if isinstance(in_dict, dict):
            return SimpleNamespace(**{k: cls.dict_to_obj(v) for k, v in in_dict.items()})
        
        raise(TypeError('Must supply dictionary as input'))
    
    
    @classmethod
    def to_dict(cls, obj, classkey=None):
        """
        Recursively translate object into dictionary format
        
        Arguments:
            obj: object to translate

        Returns:
            A dictionary representation of the object
        """
        if isinstance(obj, dict):
            data = {}
            for (k, v) in obj.items():
                data[k] = cls.to_dict(v, classkey)
            return data
        elif hasattr(obj, "_ast"):
            return cls.to_dict(obj._ast())
        elif hasattr(obj, "__iter__") and not isinstance(obj, str):
            return [cls.to_dict(v, classkey) for v in obj]
        elif hasattr(obj, "__dict__"):
            data = dict([(key, cls.to_dict(value, classkey)) 
                for key, value in obj.__dict__.items() 
                if not callable(value) and not key.startswith('_')])
            if classkey is not None and hasattr(obj, "__class__"):
                data[classkey] = obj.__class__.__name__
            return data
        else:
            return obj

# =================================================================================================
class StringHelper:

    @staticmethod
    def pad_r(text: str, length: int, pad_char: str = ' ') -> str:
        """
        Pad input text with pad character, return left justified string of specified length.

        Example::
            ```
            text = pad_r('abc', 10, pad_char='X')
            print(text) 
            'abcXXXXXXXX'
            ```
        Arguments:
            text: Input string to pad.
            length: Length of resulting string.

        Keyword Arguments:
            pad_char: String padding character (default: {' '}).

        Raises:
            ValueError: Pad character MUST be of length 1.

        Returns:
            Left justified padded string.
        """
        if len(pad_char) > 1:
            raise ValueError('Padding character should only be 1 character in length')
        
        pad_len = length - len(text)
        if pad_len > 0:
            return f'{text}{pad_char*pad_len}'
        return text    

    @staticmethod
    def pad_l(text: str, length: int, pad_char: str = ' ') -> str:
        """
        Pad input text with pad character, return right-justified string of specified length.

            Example::
        
                text = pad_l('abc', 10, pad_char='X')
                print(text) 

                output:  
                'XXXXXXXXabc'

        Arguments:
            text: Input string to pad.
            length: Length of resulting string.

        Keyword Arguments:
            pad_char: String padding character [default: {' '}].

        Raises:
            ValueError: Pad character MUST be of length 1.

        Returns:
            Right justified padded string.
        """
        if len(pad_char) > 1:
            raise ValueError('Padding character should only be 1 character in length')
        
        pad_len = length - len(text)
        if pad_len > 0:
            return f'{pad_char*pad_len}{text}'
        return text    
    
    @staticmethod
    def center(text: str, length: int, pad_char: str = ' ') -> str:
        """
        Center text in a string of size length with padding pad.

        Args:
            text (str): text to be centered.
            length (int): length of resulting string.  If it is less
              than len(text), text will be returned.
            pad (str, optional): padding character (1). Defaults to ' '.

        Returns:
            str: text centered string
        """
        if len(pad_char) > 1:
            raise ValueError('Padding character should only be 1 character in length')
        
        new_str = text
        text_len = len(text)
        if text_len < length:
            pad_len = max(int((length - text_len) / 2) + text_len, text_len+1)
            new_str = StringHelper.pad_l(text=new_str, length=pad_len, pad_char=pad_char)
            new_str = StringHelper.pad_r(text=new_str, length=length, pad_char=pad_char)

        return new_str
    
if __name__ == "__main__":
    print(StringHelper.pad_l(' pad_l', 20, '-'))
    print(StringHelper.pad_r('pad_r ', 20, '-'))
    for length in range(20, len(' center ')-1, -2):
        print(StringHelper.center(' center ', length, '-'))
    



"""
Helper for OS functions.

Supports windows and linux

"""
import ctypes
import os
import pathlib
import platform
import signal
import sys

from loguru import logger as LOGGER


class OSHelper():
    """
    Helper class for OS functions.
    
    Supports windows and linux

    Features:
        - OS/hardware detection (Windows/Linux/RaspberryPI)
        - Process detection (running in fg/bg)
        - Is executable in the path?
        - Is process running with admin/root permisssions?
        
        Windowws:
        - is_admin
        - elevate to admin

        Linux:
        - is_root

    Raises:
        OSError: On un-supported OS

    Examples::
        from dt_tools.os.os_helper import OSHelper

        print(f'Is Windows: {OSHelper.is_windows()})
        print(f'Is Linux  : {OSHelper.is_linux()})
        print(f'Is RPi    : {OSHelper.is_raspberrypi()})
        
    """
    @staticmethod
    def is_windows() -> bool:
        """Return True if running in windows else False"""
        return platform.system() == "Windows"
    
    @staticmethod
    def is_linux() -> bool:
        """Return True if running in linux else False"""
        return platform.system() == "Linux"

    @staticmethod
    def is_raspberrypi() -> bool:
        """
        Check if hardware is a Raspberry PI

        Returns:
            True if Raspberry PI else False
        """
        if not OSHelper.is_linux():
            return False
        buffer = []
        with open('/proc/cpuinfo','r') as fh:
            buffer = fh.readlines()
        token = [x for x in buffer if x.startswith('Hardware')]
        hw = token[0].split(":")[1].strip()
        if hw.startswith("BCM"):
            return True
        return False

    @staticmethod
    def is_running_in_foreground():
        """
        Check if process is running in foreground

        Returns:
            True if running in foreground else False
        """
        try:
            if os.getpgrp() == os.tcgetpgrp(sys.stdout.fileno()):
                return True     # is foreground
            return False        # is background
        except AttributeError:
            # Fall back, looks like os.getpgrp() is not available
            return sys.stdout.isatty()
        except OSError:
            return True         # is as a daemon       

    @staticmethod
    def is_running_in_background():
        """
        Check if process is running in background (or as a daemon)

        Returns:
            True if running in background else False
        """
        # return not cls.is_running_in_foreground()
        return not OSHelper.is_running_in_foreground()

    @staticmethod
    def is_executable_available(name: str) -> str:
        """
        Is executable in system path?

        Arguments:
            name: Name of executable.

        Returns:
            Fully qualified executable path if found, else None
        """
        if OSHelper.is_windows():
            sep = ';'
        else:
            sep = ':'
        PATH = os.getenv('PATH')
        exe = None
        found = False
        for dir in PATH.split(sep):
            exe = pathlib.Path(dir) / name
            if exe.exists():
                found = True
                break
            if OSHelper.is_windows():
                exe = pathlib.Path(dir) / f'{name}.exe'
                if exe.exists():
                    found = True
                    break
                exe = pathlib.Path(dir) / f'{name}.com'
                if exe.exists():
                    found = True
                    break

        if found:
            return exe
        return None

    @staticmethod
    def is_windows_admin():
        """
        Is process running as Windows Admin

        Returns:
            True if Admin privileges in effect else False
        """
        if OSHelper.is_windows():
            try:
                return ctypes.windll.shell32.IsUserAnAdmin()
            except Exception as ex:
                LOGGER.warning(f'On Windows, but cant check Admin privileges: {repr(ex)}')
                return False            
        
        return False

    @staticmethod
    def is_linux_root():
        """
        Is process running as root?

        Returns:
            True if root else False
        """
        return os.geteuid() == 0
    
    @staticmethod
    def is_god(cls):
        """
        Is process running elevated.

        For windows: admin permissions
        For linux:   root user
        
        Returns:
            True if admin/root else False
        """
        if OSHelper.is_windows:
            return OSHelper.is_windows_admin()
        return OSHelper.is_linux_root()
    
    @staticmethod
    def elevate_to_admin() -> bool:
        """
        Relaunch process with elevated privileges to Windows Admin.

        User will be presented with a prompt which must be ACK'd for elevation.

        Raises:
            OSError: If not running on Windows.

        Returns:
            bool: True if successful else False
        """
        if not OSHelper.is_windows():
            raise OSError('run_as_admin is ONLY available in Windows')

        if OSHelper.is_windows_admin():
            return True
        
        # Re-run the program with admin rights
        LOGGER.debug(f'Run Elevated - sys.executable: {sys.executable}   args: {sys.argv}')
        hresult = ctypes.windll.shell32.ShellExecuteW(None, "runas", sys.executable, " ".join(sys.argv), None, 1)
        LOGGER.debug(f'  returns {hresult}')
        return True if hresult > 32 else False

    @staticmethod
    def disable_ctrl_c_handler() -> bool:
        """
        Disable handler for Ctrl-C checking.

        Returns:
          True if successful, else False
        """
        success = True
        try:
            signal.signal(signal.SIGINT, signal.SIG_DFL)
        except:  # noqa: E722
            success = False
        return success

    @staticmethod
    def enable_ctrl_c_handler(handler_function: callable = None) -> bool:
        """
        Enable handler for Ctrl-C checking.
        
        If Ctrl-C occurs, and no handler function has been defined, user is prompted to continue or exit.

        Arguments:
            handler_function: Function to be called when ctrl-c is requested. (optional) 
              If supplied, the function should be defined as follows...
            
              Example::

                def handler_name(signum, frame):
                    code to execute when handler is called...  

        Returns:
            True if handler successfully enabled, else False.

        """
        success = True
        if handler_function is None:
            handler_function = OSHelper._interrupt_handler
            
        try:
            signal.signal(signal.SIGINT, handler_function)
        except:  # noqa: E722
            success = False
        return success

    @staticmethod
    def _interrupt_handler(signum, frame):
        resp = ''
        while resp not in ['c', 'e']:
            resp = input('\nCtrl-C, Continue or Exit (c,e)? ').lower()
            if resp == 'e':
                os._exit(1)

OSHelper.enable_ctrl_c_handler()
if __name__ == "__main__":
    import dt_tools.cli.dt_misc_os_demo as module
    module.demo()